/**
 * display.c
 *
 * by YOSHIDA Kazuhiro (moriq.kazuhiro@nifty.ne.jp)
 */

#include "ruby.h"
#include "init.h"

#include <locale.h>
#include <X11/keysym.h>

#include "display.h"
#include "window.h"
#include "font.h"
#include "gc.h"
#include "event.h"
#include "color.h"
#include "event.h"
#include "debug.h"

void
xlib_no_display()
{
	rb_raise(rb_eRuntimeError, "already closed display");
}

void
xlib_free_ic(ic)
	XIC ic;
{
	XDestroyIC(ic);
}

static void
xlib_free_window_list(disp)
	t_display *disp;
{
	t_window *node;

	Debug("free window list begin");
	node = disp->window_list_head->next;
	while (node != disp->window_list_tail) {
		if (node->ic) {
			xlib_free_ic(node->ic);
			node->ic = 0;
		}
		if (node->xwindow) {
			if (node->xparent)
				XDestroyWindow(node->xdisplay, node->xwindow);
			else
				   XFreePixmap(node->xdisplay, node->xwindow);
			node->xwindow = 0;
		}
		node->next->prev = node->prev;
		node->prev->next = node->next;
		node = node->next;
	}
	Debug("free window list end");
}

static void
xlib_free_font_list(disp)
	t_display *disp;
{
	t_font *node;

	Debug("free font list begin");
	node = disp->font_list_head->next;
	while (node != disp->font_list_tail) {
		if (node->xfont) {
			XFreeFontSet(disp->xdisplay, node->xfont);
			node->xfont = 0;
		}
		node->next->prev = node->prev;
		node->prev->next = node->next;
		node = node->next;
	}
	Debug("free font list end");
}

static void
xlib_free_gc_list(disp)
	t_display *disp;
{
	t_gc *node;

	Debug("free gc list begin");
	node = disp->gc_list_head->next;
	while (node != disp->gc_list_tail) {
		if (node->xgc) {
			XFreeGC(disp->xdisplay, node->xgc);
			node->xgc = 0;
		}
		node->next->prev = node->prev;
		node->prev->next = node->next;
		node = node->next;
	}
	Debug("free gc list end");
}

static void
x_close_display(disp)
	t_display *disp;
{
#ifdef HAVE_IMLIB_H
	if (disp->idp) {
		im_free_data(disp->idp);
		disp->idp = NULL;
	}
#endif
	if (disp->xdisplay) {
		xlib_free_window_list(disp);
		xlib_free_font_list(disp);
		xlib_free_gc_list(disp);

		if (disp->im) {
			XCloseIM(disp->im);
			disp->im = NULL;
		}
		XCloseDisplay(disp->xdisplay);
		disp->xdisplay = NULL;

		free(disp->window_list_head);
		free(disp->window_list_tail);
		free(disp->gc_list_head);
		free(disp->gc_list_tail);
		free(disp->font_list_head);
		free(disp->font_list_tail);
		free(disp->visual);
	}
}

static void
xlib_free_display(disp)
	t_display *disp;
{
	Debug("free display begin");
	x_close_display(disp);
	free(disp);
	Debug("free display end");
}

static VALUE
xlib_display_close(self)
	VALUE self;
{
	t_display *disp;
	Debug("close display begin");
	GetDisplay(self, disp);
	x_close_display(disp);
	Debug("close display end");
	return Qnil;
}

static VALUE
xlib_display_new(argc, argv, self)
	int argc;
	VALUE *argv;
	VALUE self;
{
	t_display *disp;
	Display *xdisplay;
	VALUE obj, arg;
	
	t_window *winp;
	t_font *fonp;
	t_gc *gcp;
	t_visual vinfo;
	VALUE root, font, gc;
	int xscreen, xwindow, num, i;

	if (rb_scan_args(argc, argv, "01", &arg) == 0 || NIL_P(arg))
		xdisplay = XOpenDisplay(NULL);
	else
		xdisplay = XOpenDisplay(STR2CSTR(arg));

	if (xdisplay == NULL) return Qnil;

	xscreen = DefaultScreen(xdisplay);
	xwindow = DefaultRootWindow(xdisplay);
	vinfo.screen = xscreen;

	obj = Data_Make_Struct(self, t_display, 0, xlib_free_display, disp);
	disp->window_list_head = ALLOC(t_window);
	disp->window_list_tail = ALLOC(t_window);
	disp->window_list_head->prev = disp->window_list_head;
	disp->window_list_head->next = disp->window_list_tail;
	disp->window_list_tail->prev = disp->window_list_head;
	disp->window_list_tail->next = disp->window_list_tail;

	disp->font_list_head = ALLOC(t_font);
	disp->font_list_tail = ALLOC(t_font);
	disp->font_list_head->prev = disp->font_list_head;
	disp->font_list_head->next = disp->font_list_tail;
	disp->font_list_tail->prev = disp->font_list_head;
	disp->font_list_tail->next = disp->font_list_tail;

	disp->gc_list_head = ALLOC(t_gc);
	disp->gc_list_tail = ALLOC(t_gc);
	disp->gc_list_head->prev = disp->gc_list_head;
	disp->gc_list_head->next = disp->gc_list_tail;
	disp->gc_list_tail->prev = disp->gc_list_head;
	disp->gc_list_tail->next = disp->gc_list_tail;

	disp->xdisplay = xdisplay;
	disp->im = XOpenIM(xdisplay, NULL, NULL, NULL);
	disp->proto = XInternAtom(xdisplay, "WM_DELETE_WINDOW", True);
#ifdef HAVE_IMLIB_H
	disp->idp = NULL;
#endif
	disp->visual = XGetVisualInfo(xdisplay, VisualScreenMask, &vinfo, &num);
	{
		VALUE ary, visual;

		/* free disp->visual array at xlib_display_free() */
		ary = rb_ary_new2(num);
		for (i=0; i<num; i++) {
			visual = Data_Wrap_Struct(cXlib_visual, 0, 0, &disp->visual[i]);
			rb_ary_store(ary, i, visual);
		}
		rb_iv_set(obj, "@visuals", ary);
	}

	winp = disp->window_list_head;
	root = Data_Wrap_Struct(cXlib_window, 0, 0, winp);
	winp->xdisplay = xdisplay;
	winp->xparent = xwindow;
	winp->xwindow = xwindow;
	winp->ic = 0;
	rb_iv_set(root, "@display", obj);
	rb_iv_set(root, "@parent", root);
	rb_iv_set(root, "@shown", Qtrue);
	rb_iv_set(root, "@events", rb_hash_new());

	fonp = disp->font_list_head;
	font = Data_Wrap_Struct(cXlib_font, 0, 0, fonp);
	fonp->xdisplay = xdisplay;
	fonp->xfont = xlib_font_load(xdisplay, DEFAULT_XFONT);
	fonp->nfont = DEFAULT_NFONT;

	gcp = disp->gc_list_head;
	gc = Data_Wrap_Struct(cXlib_gc, 0, 0, gcp);
	gcp->xgc = DefaultGC(xdisplay, xscreen);
	gcp->xdisplay = xdisplay;
	gcp->xwindow = xwindow;
	rb_iv_set(gc, "@display", obj);
	rb_iv_set(gc, "@window", root);
	{
		VALUE ary;

		ary = rb_ary_new();
		rb_ary_push(ary, root);
		rb_iv_set(obj, "@windows", ary);
	}
	rb_iv_set(obj, "@root", root);
	rb_iv_set(obj, "@font", font);
	rb_iv_set(obj, "@gc", gc);

	rb_obj_call_init(obj, 0, NULL);

	return obj;
}

static VALUE
xlib_display_flush(self)
	VALUE self;
{
	t_display *disp;
	GetDisplay(self, disp);
	XFlush(disp->xdisplay);
	return Qnil;
}

static VALUE
xlib_display_sync(self)
	VALUE self;
{
	t_display *disp;
	GetDisplay(self, disp);
	XSync(disp->xdisplay, False);
	return Qnil;
}

static VALUE
xlib_display_alloc_color(self, arg)
	VALUE self, arg;
{
	t_display *disp;
	Display *xdisplay;
	t_color *colp;
	unsigned long pixel;
	Colormap cmap;
	XColor color, exact;
	
	GetDisplay(self, disp);
	xdisplay = disp->xdisplay;
	cmap = DefaultColormap(xdisplay, 0);
	switch (TYPE(arg)) {
	case T_STRING:
		XAllocNamedColor(xdisplay, cmap, RSTRING(arg)->ptr, &color, &exact);
		pixel = color.pixel;
		break;
	default:
		Data_Get_Struct(arg, t_color, colp);
		color = *colp;
		XAllocColor(xdisplay, cmap, &color);
		pixel = color.pixel;
		break;
	}
	return INT2NUM(pixel);
}

static VALUE
xlib_display_alloc_color_cells(argc, argv, self)
	int argc;
	VALUE *argv;
	VALUE self;
{
	t_display *disp;
	Display *xdisplay;
	Colormap cmap;
	unsigned long *pix_base, *sub_mask;
	int base_len, mask_len;
	Bool result;
	unsigned int i;
	VALUE base, mask;
	
	if (rb_scan_args(argc, argv, "11", &base, &mask) == 1 || NIL_P(mask)) {
		mask_len = 0;
		sub_mask = NULL;
	} else {
		Check_Type(mask, T_ARRAY);
		mask_len = RARRAY(mask)->len;
		sub_mask = ALLOC_N(unsigned long, mask_len);
	}
	Check_Type(base, T_ARRAY);
	base_len = RARRAY(base)->len;
	pix_base = ALLOC_N(unsigned long, base_len);

	GetDisplay(self, disp);
	xdisplay = disp->xdisplay;
	cmap = DefaultColormap(xdisplay, DefaultScreen(xdisplay));
	result = XAllocColorCells(xdisplay, cmap, True, sub_mask, mask_len, pix_base, base_len);
	printf("%d %d %d\n", result, mask_len, base_len);
	
	for (i=0; i<mask_len; i++) RARRAY(mask)->ptr[i] = INT2NUM(sub_mask[i]);
	for (i=0; i<base_len; i++) RARRAY(base)->ptr[i] = INT2NUM(pix_base[i]);

	if (sub_mask) free(sub_mask);
	if (pix_base) free(pix_base);
	
	return (result ? Qtrue : Qfalse);
}

static VALUE
xlib_display_store_color(argc, argv, self)
	int argc;
	VALUE *argv;
	VALUE self;
{
	t_display *disp;
	Display *xdisplay;
	t_color *colp;
	Colormap cmap;
	XColor color;
	XColor *colors;
	unsigned long flags;
	int i, num;
	VALUE top, pix;

	GetDisplay(self, disp);
	xdisplay = disp->xdisplay;
	cmap = DefaultColormap(xdisplay, 0);

	if (rb_scan_args(argc, argv, "11", &top, &pix) == 1) {
		switch ( TYPE(top) ) {
		case T_ARRAY:
			puts("XStoreColors");
			num = RARRAY(top)->len;
			colors = ALLOC_N(XColor, num);
			for (i=0; i<num; i++) {
				Data_Get_Struct(RARRAY(top)->ptr[i], t_color, colp);
				colors[i] = *colp;
			}
			XStoreColors(xdisplay, cmap, colors, num);
			free(colors);
			break;
		default:
			puts("XStoreColor");
			Data_Get_Struct(top, t_color, colp);
			color = *colp;
			XStoreColor(xdisplay, cmap, &color);
			break;
		}
	} else {
		puts("XStoreNamedColor");
		Check_Type(top, T_STRING);
		flags = DoGreen | DoRed | DoBlue;
		XStoreNamedColor(xdisplay, cmap, RSTRING(top)->ptr, NUM2INT(pix), flags);
	}
	return Qnil;
}

static VALUE
xlib_display_get_depth(self)
	VALUE self;
{
	t_display *disp;
	Display *x;

	GetDisplay(self, disp);
	x = disp->xdisplay;

	return INT2NUM(DefaultDepth(x, DefaultScreen(x)));
}

void
xlib_init_display()
{
	cXlib_display = rb_define_class_under(mXlib, "Display", rb_cObject);
	DefineSingletonMethod(display, new, -1)
	DefineMethod(display, close, 0)
	DefineMethod(display, flush, 0)
	DefineMethod(display, sync, 0)
	DefineMethod(display, alloc_color, 1)
	DefineMethod(display, alloc_color_cells, -1)
	DefineMethod(display, store_color, -1)
	DefineMethodGet(display, depth)
	rb_define_attr(cXlib_display, "windows", 1, 0);
	rb_define_attr(cXlib_display, "root", 1, 0);
	rb_define_attr(cXlib_display, "font", 1, 0);
	rb_define_attr(cXlib_display, "gc", 1, 0);
	rb_define_attr(cXlib_display, "visuals", 1, 0);
}
