/**
 * event.c
 *
 * by YOSHIDA Kazuhiro (moriq.kazuhiro@nifty.ne.jp)
 */

#include "ruby.h"
#include "init.h"

#include <locale.h>
#include <X11/keysym.h>

#include "display.h"
#include "window.h"
#include "rect.h"
#include "event.h"
#include "debug.h"

unsigned long events =
	  ExposureMask
	| ButtonPressMask
	| ButtonReleaseMask
	| EnterWindowMask
	| LeaveWindowMask
	| KeyPressMask
	| KeyReleaseMask
	| PointerMotionMask
	| StructureNotifyMask
	;

DefIntern(delete)
DefIntern(exposed)
DefIntern(button_press)
DefIntern(button_release)
DefIntern(enter_notify)
DefIntern(leave_notify)
DefIntern(key_press)
DefIntern(key_release)
DefIntern(motion_notify)
DefIntern(configure)

static VALUE sXlib_ExposeEvent;

VALUE
xlib_expose_event_setup(src, ev)
	VALUE src;
	XExposeEvent *ev;
{
	return rb_struct_new(sXlib_ExposeEvent,
/*
		INT2FIX(ev->type),
		INT2NUM(ev->serial),
		(ev->send_event) ? Qtrue : Qfalse,
		src,
*/
		rb_funcall(cXlib_rect, rb_intern("new"), 4,
			INT2NUM(ev->x),
			INT2NUM(ev->y),
			INT2NUM(ev->width ),
			INT2NUM(ev->height)
		),
		INT2FIX(ev->count),
/*
		INT2FIX(ev->x_root),
		INT2FIX(ev->y_root),
		INT2FIX(ev->state),
		(ev->same_screen) ? Qtrue : Qfalse,
*/
		0 );
}

static void
xlib_init_expose_event()
{
	sXlib_ExposeEvent = rb_struct_define("ExposeEvent",
/*
		"type",
		"serial",
		"send_p",
		"source",
*/
		"rect",
		"count",
/*
		"x_root",
		"y_root",
		"state",
		"same_screen_p",
*/
		0 );
	rb_global_variable(&sXlib_ExposeEvent);
}

static VALUE sXlib_ButtonEvent;

VALUE
xlib_button_event_setup(src, ev)
	VALUE src;
	XButtonEvent *ev;
{
	return rb_struct_new(sXlib_ButtonEvent,
/*
		INT2FIX(ev->type),
		INT2NUM(ev->serial),
		(ev->send_event) ? Qtrue : Qfalse,
		src,
*/
		INT2NUM(ev->time),
		INT2FIX(ev->x),
		INT2FIX(ev->y),
		INT2FIX(ev->button),
/*
		INT2FIX(ev->x_root),
		INT2FIX(ev->y_root),
		INT2FIX(ev->state),
		(ev->same_screen) ? Qtrue : Qfalse,
*/
		0 );
}

static void
xlib_init_button_event()
{
	sXlib_ButtonEvent = rb_struct_define("ButtonEvent",
/*
		"type",
		"serial",
		"send_p",
		"source",
*/
		"time",
		"x",
		"y",
		"button",
/*
		"x_root",
		"y_root",
		"state",
		"same_screen_p",
*/
		0 );
	rb_global_variable(&sXlib_ButtonEvent);
}

static VALUE sXlib_KeyEvent;

VALUE
xlib_key_event_setup(src, ev)
	VALUE src;
	XKeyEvent *ev;
{
	t_window *winp;
	KeySym keysym;
	VALUE str = Qnil;

	GetWindow(src, winp);

	if (!winp->ic) {
		keysym = XLookupKeysym(ev, 0);
	} else {
		char buffer[256]; /* enough */
		Status status;
		int num;

		num = XmbLookupString(winp->ic, ev,
			buffer, sizeof(buffer), &keysym, &status);
		switch(status) {
		case XBufferOverflow:
			rb_raise(rb_eRuntimeError, "buffer overflow");
		case XLookupNone:
			break;
		case XLookupChars:
			str = rb_str_new(buffer, num);
			break;
		case XLookupKeySym:
		case XLookupBoth:
			if (0 <= keysym && keysym <= 0xff) {
				char c = keysym;
				str = rb_str_new(&c, 1);
			}
			break;
		default:
			rb_raise(rb_eRuntimeError, "unknown status");
		}
	}
	return rb_struct_new(sXlib_KeyEvent,
/*
		INT2FIX(ev->type),
		INT2NUM(ev->serial),
		(ev->send_event) ? Qtrue : Qfalse,
		src,
*/
		INT2NUM(ev->time),
		INT2FIX(ev->x),
		INT2FIX(ev->y),
		INT2NUM(ev->keycode),
		INT2NUM(keysym),
		str,
/*
		INT2FIX(ev->x_root),
		INT2FIX(ev->y_root),
		INT2FIX(ev->state),
		(ev->same_screen) ? Qtrue : Qfalse,
*/
		0 );
}

static void
xlib_init_key_event()
{
	sXlib_KeyEvent = rb_struct_define("KeyEvent",
/*
		"type",
		"serial",
		"send_p",
		"source",
*/
		"time",
		"x",
		"y",
		"keycode",
		"keysym",
		"string",
/*
		"x_root",
		"y_root",
		"state",
		"same_screen_p",
*/
		0 );
	rb_global_variable(&sXlib_KeyEvent);
}

static VALUE sXlib_MotionEvent;

VALUE
xlib_motion_event_setup(src, ev)
	VALUE src;
	XMotionEvent *ev;
{
	return rb_struct_new(sXlib_MotionEvent,
/*
		INT2FIX(ev->type),
		INT2NUM(ev->serial),
		(ev->send_event) ? Qtrue : Qfalse,
		src,
*/
		INT2NUM(ev->time),
		INT2FIX(ev->x),
		INT2FIX(ev->y),
		(ev->is_hint) ? Qtrue : Qfalse,
/*
		INT2FIX(ev->x_root),
		INT2FIX(ev->y_root),
		INT2FIX(ev->state),
		(ev->same_screen) ? Qtrue : Qfalse,
*/
		0 );
}

static void
xlib_init_motion_event()
{
	sXlib_MotionEvent = rb_struct_define("MotionEvent",
/*
		"type",
		"serial",
		"send_p",
		"source",
*/
		"time",
		"x",
		"y",
		"is_hint",
/*
		"x_root",
		"y_root",
		"state",
		"same_screen_p",
*/
		0 );
	rb_global_variable(&sXlib_MotionEvent);
}

static VALUE sXlib_CrossingEvent;

VALUE
xlib_crossing_event_setup(src, ev)
	VALUE src;
	XCrossingEvent *ev;
{
	return rb_struct_new(sXlib_CrossingEvent,
/*
		INT2FIX(ev->type),
		INT2NUM(ev->serial),
		(ev->send_event) ? Qtrue : Qfalse,
		src,
*/
		INT2NUM(ev->time),
		INT2FIX(ev->x),
		INT2FIX(ev->y),
/*
		INT2FIX(ev->x_root),
		INT2FIX(ev->y_root),
		INT2FIX(ev->state),
		(ev->same_screen) ? Qtrue : Qfalse,
*/
		0 );
}

static void
xlib_init_crossing_event()
{
	sXlib_CrossingEvent = rb_struct_define("CrossingEvent",
/*
		"type",
		"serial",
		"send_p",
		"source",
*/
		"time",
		"x",
		"y",
/*
		"x_root",
		"y_root",
		"state",
		"same_screen_p",
*/
		0 );
	rb_global_variable(&sXlib_CrossingEvent);
}

static VALUE sXlib_ConfigureEvent;

VALUE
xlib_configure_event_setup(src, ev)
	VALUE src;
	XConfigureEvent *ev;
{
	return rb_struct_new(sXlib_ConfigureEvent,
/*
		INT2FIX(ev->type),
		INT2NUM(ev->serial),
		(ev->send_event) ? Qtrue : Qfalse,
		src,
*/
		rb_funcall(cXlib_rect, rb_intern("new"), 4,
			INT2NUM(ev->x),
			INT2NUM(ev->y),
			INT2NUM(ev->width ),
			INT2NUM(ev->height)
		),
		INT2FIX(ev->border_width),
/*
		INT2FIX(ev->x_root),
		INT2FIX(ev->y_root),
		INT2FIX(ev->state),
		(ev->same_screen) ? Qtrue : Qfalse,
*/
		0 );
}

static void
xlib_init_configure_event()
{
	sXlib_ConfigureEvent = rb_struct_define("ConfigureEvent",
/*
		"type",
		"serial",
		"send_p",
		"source",
*/
		"rect",
		"border_width",
/*
		"x_root",
		"y_root",
		"state",
		"same_screen_p",
*/
		0 );
	rb_global_variable(&sXlib_ConfigureEvent);
}

VALUE
xlib_event_dummy(self, args)
	VALUE self, args;
{
	return Qnil;
}

void
xlib_event(self, id, args)
	VALUE self, args;
	ID id;
{
	VALUE hash, ary, key;
	 
	hash = rb_iv_get(self, "@events");
	key = INT2FIX(id);
	ary = rb_hash_aref(hash, key);
	if (ary != RHASH(hash)->ifnone) {
	 	int i;
	 	for (i=0; i<RARRAY(ary)->len; i++) {
			rb_apply(RARRAY(ary)->ptr[i], id_call, args);
		}
	}
	rb_apply(self, id, args);
}

static VALUE
xlib_display_event(argc, argv, self)
	int    argc;
	VALUE *argv;
	VALUE  self;
{
	XEvent ev;
	t_display *disp;
	t_window *winp = NULL;
	ID id = 0;
	VALUE grab, args = Qnil, act = Qnil, ret = Qtrue;

	GetDisplay(self, disp);
	if (!XPending(disp->xdisplay)) return ret;
	XNextEvent(disp->xdisplay, &ev);
	if (XFilterEvent(&ev, 0) == True) return ret;

	if (rb_scan_args(argc, argv, "01", &grab) == 1) {
		Window xwindow = ev.xany.window;
		t_window *grab_winp;

		GetWindow(grab, grab_winp);
		if (grab_winp->xwindow == xwindow)
			act = grab;
		else
			return ret;
	} else {
		int i;
		Window xwindow = ev.xany.window;
		VALUE ary, window;

		ary = rb_iv_get(self, "@windows");
		Check_Type(ary, T_ARRAY);

		for (i=0; i<RARRAY(ary)->len; i++) {
			window = RARRAY(ary)->ptr[i];
			Data_Get_Struct(window, t_window, winp);
			if (winp->xwindow == xwindow) {
				act = window;
				break;
			}
		}
	}
	if (act == Qnil) return ret;

	switch (ev.type) {
	case Expose:
		id = id_exposed;
		args = rb_ary_new3( 1, xlib_expose_event_setup(act, (XExposeEvent *)&ev) );
		break;
	case ButtonPress:
		id = id_button_press;
		args = rb_ary_new3( 1, xlib_button_event_setup(act, (XButtonEvent *)&ev) );
		break;
	case ButtonRelease:
		id = id_button_release;
		args = rb_ary_new3( 1, xlib_button_event_setup(act, (XButtonEvent *)&ev) );
		break;
	case EnterNotify:
		id = id_enter_notify;
		args = rb_ary_new3( 1, xlib_crossing_event_setup(act, (XCrossingEvent *)&ev) );
		break;
	case LeaveNotify:
		id = id_leave_notify;
		args = rb_ary_new3( 1, xlib_crossing_event_setup(act, (XCrossingEvent *)&ev) );
		break;
	case KeyPress:
		id = id_key_press;
		args = rb_ary_new3( 1, xlib_key_event_setup(act, (XKeyEvent *)&ev) );
		break;
	case KeyRelease:
		id = id_key_release;
		args = rb_ary_new3( 1, xlib_key_event_setup(act, (XKeyEvent *)&ev) );
		break;
	case MotionNotify:
		id = id_motion_notify;
		args = rb_ary_new3( 1, xlib_motion_event_setup(act, (XMotionEvent *)&ev) );
		break;
	case FocusIn:
		Debug("focus in");
		if (winp->ic) XSetICFocus(winp->ic);
		return ret;
	case FocusOut:
		Debug("focus out");
		if (winp->ic) XUnsetICFocus(winp->ic);
		return ret;
	case ConfigureNotify:
		id = id_configure;
		args = rb_ary_new3( 1, xlib_configure_event_setup(act, (XConfigureEvent *)&ev) );
		break;
	case ClientMessage:
		if (ev.xclient.data.l[0] != disp->proto) break;
		id = id_delete;
		args = rb_ary_new2(0);
		ret = Qfalse;
		break;
	default:
		return ret;
	}
	xlib_event(act, id, args);
	return ret;
}

void
xlib_init_event()
{
	 SetIntern(delete)
	 SetIntern(exposed)
	 SetIntern(button_press)
	 SetIntern(button_release)
	 SetIntern(enter_notify)
	 SetIntern(leave_notify)
	 SetIntern(key_press)
	 SetIntern(key_release)
	 SetIntern(motion_notify)
	 SetIntern(configure)

	 xlib_init_expose_event();
	 xlib_init_button_event();
	 xlib_init_key_event();
	 xlib_init_motion_event();
	 xlib_init_crossing_event();
	 xlib_init_configure_event();

	 DefineMethod(display, event, -1)
}
