/**
 * font.c
 */

#include "ruby.h"
#include "init.h"

#include "display.h"
#include "window.h"
#include "font.h"
#include "gc.h"
#include "rect.h"
#include "debug.h"

void
xlib_no_font()
{
	rb_raise(rb_eRuntimeError, "already closed font");
}

static void
xlib_free_font(fonp)
	t_font *fonp;
{
	Debug("xlib free font begin");
	if (fonp->xfont) {
		XFreeFontSet(fonp->xdisplay, fonp->xfont);
		fonp->xfont = 0;
	}
	fonp->next->prev = fonp->prev;
	fonp->prev->next = fonp->next;
	free(fonp);
	Debug("xlib free font end");
}

static VALUE
xlib_font_close(self)
	VALUE self;
{
	t_font *fonp;
	Debug("xlib font close begin");
	GetFont(self, fonp);
	XFreeFontSet(fonp->xdisplay, fonp->xfont);
	fonp->xfont = 0;
	Debug("xlib font close end");
	return Qnil;
}

XFontSet
xlib_font_load(xdisplay, font_name)
	Display *xdisplay;
	char *font_name;
{
	XFontSet fontset;
	int missing_charset_count;
	char **missing_charset_list;
	char *def_string;

	fontset = XCreateFontSet(xdisplay, font_name,
		&missing_charset_list, &missing_charset_count,
		&def_string);
	if (missing_charset_count) {
		int i;
		fprintf(stderr, "missing charsets in Xlib::Font.new\n");
		for (i=0; i<missing_charset_count; i++)
			fprintf(stderr, "  %s\n", missing_charset_list[i]);
		XFreeStringList(missing_charset_list);
	}
	return fontset;
}

static VALUE
xlib_font_create(argc, argv, self, display)
	int    argc;
	VALUE *argv;
	VALUE  self, display;
{
	XFontSet xfont;
	t_display *disp;
	t_font *fonp;
	VALUE obj, font_list, font_name;

	GetDisplay(display, disp);

	xlib_raise_if_argc_none(argc, "need Array");
	font_list = *argv;
	Check_Type(font_list, T_ARRAY);
	if (--argc) ++argv;

	xlib_raise_if_argc_rest(argc);

	font_name = rb_ary_join(font_list, rb_str_new2(","));

	xfont = xlib_font_load(disp->xdisplay, STR2CSTR(font_name));
	if (!xfont) return Qnil;

	obj = Data_Make_Struct(self, t_font, 0, xlib_free_font, fonp);
	fonp->prev = disp->font_list_head;
	fonp->next = disp->font_list_head->next;
	disp->font_list_head->next->prev = fonp;
	disp->font_list_head->next = fonp;
	fonp->xdisplay = disp->xdisplay;
	fonp->xfont = xfont;
	fonp->nfont = RARRAY(font_list)->len;

	return obj;
}

static VALUE
xlib_font_new(argc, argv, self)
	int    argc;
	VALUE *argv;
	VALUE  self;
{
	VALUE display;

	xlib_raise_if_argc_none(argc, "need Display");
	display = *argv;
	if (--argc) ++argv;

	return xlib_font_create(argc, argv, self, display);
}

static VALUE
xlib_font_escapement(self, str)
	VALUE self, str;
{
	t_font *fonp;
	int width;

	GetFont(self, fonp);
	Check_Type(str, T_STRING);

	width = XmbTextEscapement(fonp->xfont,
		RSTRING(str)->ptr,
		RSTRING(str)->len);

	return INT2NUM(width);
}

static VALUE
xlib_font_extents(self, str)
	VALUE self, str;
{
	t_font *fonp;
	t_rect *_ink, *_log;
	VALUE ink, log;

	GetFont(self, fonp);
	Check_Type(str, T_STRING);

	ink = Data_Make_Struct(cXlib_rect, t_rect, 0, xlib_free_rect, _ink);
	log = Data_Make_Struct(cXlib_rect, t_rect, 0, xlib_free_rect, _log);

	XmbTextExtents(fonp->xfont,
		RSTRING(str)->ptr,
		RSTRING(str)->len,
		_ink,
		_log);

	return rb_assoc_new(ink, log);
}

static VALUE
xlib_font_id_name(self)
{
	 t_font *fonp;
	 XFontStruct **data;
	 char **name;
	 VALUE hash;
	 int i;

	 GetFont(self, fonp);
	 XFontsOfFontSet(fonp->xfont, &data, &name);

	 hash = rb_hash_new();
	 for (i=0; i<fonp->nfont; i++) {
			rb_hash_aset(hash, INT2NUM(data[i]->fid), rb_str_new2(name[i]));
	 }
	 return hash;
}

static VALUE
xlib_display_new_font(argc, argv, self)
	int    argc;
	VALUE *argv;
	VALUE  self;
{
	return xlib_font_create(argc, argv, cXlib_font, self);
}

void
xlib_init_font()
{
	cXlib_font = rb_define_class_under(mXlib, "Font", rb_cObject);
	rb_define_singleton_method(cXlib_font, "new", xlib_font_new, -1);
	DefineMethod(font, close, 0)
	DefineMethod(font, escapement, 1)
	DefineMethod(font, extents, 1)
	DefineMethod(font, id_name, 0)
	/*
	DefineMethod(font, base_name, 0)
	DefineMethod(font, locale, 0)
	DefineMethodIs(font, depend_context_or_direction)
	DefineMethodIs(font, depend_context)
	DefineMethodIs(font, depend_direction)
	*/
	DefineMethod(display, new_font, -1)
}
