/**
 * gc.c
 *
 * by YOSHIDA Kazuhiro (moriq.kazuhiro@nifty.ne.jp)
 */

#include "ruby.h"
#include "init.h"

#include <locale.h>
#include <X11/keysym.h>

#include "display.h"
#include "window.h"
#include "font.h"
#include "gc.h"
#include "point.h"
#include "rect.h"
#include "debug.h"

static unsigned long
xlib_alloc_named_color(xdisplay, color_name)
	Display *xdisplay;
	char *color_name;
{
	Colormap cmap;
	XColor color, exact;

	cmap = DefaultColormap(xdisplay, DefaultScreen(xdisplay));
	XAllocNamedColor(xdisplay, cmap, color_name, &color, &exact);
	return color.pixel;
}

void
xlib_no_gc()
{
	rb_raise(rb_eRuntimeError, "already closed graphic context");
}

static void
xlib_free_gc(gcp)
	t_gc *gcp;
{
	Debug("xlib free gc begin");
	if (gcp->xgc) {
		XFreeGC(gcp->xdisplay, gcp->xgc);
		gcp->xgc = 0;
	}
	gcp->next->prev = gcp->prev;
	gcp->prev->next = gcp->next;
	free(gcp);
	Debug("xlib free gc end");
}

static VALUE
xlib_gc_close(self)
	VALUE self;
{
	t_gc *gcp;
	Debug("xlib gc close begin");
	GetGC(self, gcp);
	XFreeGC(gcp->xdisplay, gcp->xgc);
	gcp->xgc = 0;
	Debug("xlib gc close end");
	return Qnil;
}

VALUE
xlib_gc_create(argc, argv, self, window)
	int    argc;
	VALUE *argv;
	VALUE  self, window;
{
	t_display *disp;
	t_window *winp;
	t_gc *gcp;
	Display *xdisplay;
	Window xwindow;
	GC xgc;
	VALUE obj, display;
	
	display = rb_iv_get(window, "@display");
	GetDisplay(display, disp);
	GetWindow(window, winp);
	xdisplay = winp->xdisplay;
	xwindow = winp->xwindow;

	xgc = XCreateGC(xdisplay, xwindow, 0, 0);
	XSetGraphicsExposures(xdisplay, xgc, False);
	XSetForeground(xdisplay, xgc, xlib_alloc_named_color(xdisplay, "green"));

	obj = Data_Make_Struct(self, t_gc, 0, xlib_free_gc, gcp);
	gcp->prev = disp->gc_list_head;
	gcp->next = disp->gc_list_head->next;
	disp->gc_list_head->next->prev = gcp;
	disp->gc_list_head->next = gcp;
	gcp->xgc = xgc;
	gcp->xdisplay = xdisplay;
	gcp->xwindow = xwindow;
	rb_iv_set(obj, "@display", display);
	rb_iv_set(obj, "@window", window);

	return obj;
}

static VALUE
xlib_gc_new(argc, argv, self)
	int    argc;
	VALUE *argv;
	VALUE  self;
{
	VALUE window;
	int    argc_orig;
	VALUE *argv_orig;
	VALUE obj;

	argc_orig = argc;
	argv_orig = argv;

	xlib_raise_if_argc_none(argc, "need Window");
	window = *argv;
	if (--argc) ++argv;

	obj = xlib_gc_create(argc, argv, self, window);
	rb_funcall2(obj, id_init, 0, NULL);

	return obj;
}

/*
	gc attr

	gc value mask name define at X.h
	gc value struct member define at Xlib.h
	Ruby-Xlib method ident is the same as struct member ident.
*/

/**
 ** bool
 **/

#define GCDefineBooleanAttr(attr, mask)\
static VALUE \
xlib_gc_get_##attr(self)\
	VALUE self;\
{\
	t_gc *gcp;\
	XGCValues values;\
\
	GetGC(self, gcp);\
	XGetGCValues(gcp->xdisplay, gcp->xgc, mask, &values);\
	return (values.attr != 0) ? Qtrue : Qfalse;\
}\
\
static VALUE \
xlib_gc_set_##attr(self, bool)\
	VALUE self, bool;\
{\
	t_gc *gcp;\
	XGCValues values;\
\
	values.attr = ! (bool == Qnil || bool == Qfalse);\
	GetGC(self, gcp);\
	XChangeGC(gcp->xdisplay, gcp->xgc, mask, &values);\
	return bool;\
}

/**
 ** numeric
 **/

#define GCDefineNumericAttr(attr, mask)\
static VALUE \
xlib_gc_get_##attr(self)\
	VALUE self;\
{\
	t_gc *gcp;\
	XGCValues values;\
\
	GetGC(self, gcp);\
	XGetGCValues(gcp->xdisplay, gcp->xgc, mask, &values);\
	return INT2NUM(values.attr);\
}\
\
static VALUE \
xlib_gc_set_##attr(self, num)\
	VALUE self, num;\
{\
	t_gc *gcp;\
	XGCValues values;\
\
	values.attr = NUM2INT(num);\
	GetGC(self, gcp);\
	XChangeGC(gcp->xdisplay, gcp->xgc, mask, &values);\
	return num;\
}

/**
 ** pixmap
 **/

#define GCDefinePixmapAttr(attr, mask)\
static VALUE \
xlib_gc_get_##attr(self)\
	VALUE self;\
{\
	t_gc *gcp;\
	XGCValues values;\
	VALUE window;\
\
	GetGC(self, gcp);\
	XGetGCValues(gcp->xdisplay, gcp->xgc, mask, &values);\
	window = rb_iv_get(self, "@window");\
	return xlib_pixmap_setup(window, values.attr, -1, -1);\
}\
\
static VALUE \
xlib_gc_set_##attr(self, pmap)\
	VALUE self, pmap;\
{\
	t_gc *gcp;\
	XGCValues values;\
	t_window *pixp;\
\
	GetGC(self, gcp);\
	GetWindow(pmap, pixp);\
	values.attr = pixp->xwindow;\
	XChangeGC(gcp->xdisplay, gcp->xgc, mask, &values);\
	return pmap;\
}

/**
 ** setup
 **/

#define GCSetupAttr(attr)\
DefineMethodGet(gc, attr)\
DefineMethodSet(gc, attr)

/**
 ** alias
 **/

#define GCAliasAttr(new, old)\
rb_define_alias(cXlib_gc, #new, #old);\
rb_define_alias(cXlib_gc, #new "=", #old "=");

/**
 ** implement
 **/

/* logical operation */
GCDefineNumericAttr(function, GCFunction)

/* plane mask */
GCDefineNumericAttr(plane_mask, GCPlaneMask)

/* foreground pixel */
GCDefineNumericAttr(foreground, GCForeground)

/* background pixel */
GCDefineNumericAttr(background, GCBackground)

/* line width */
GCDefineNumericAttr(line_width, GCLineWidth)

/* LineSolid, LineOnOffDash, LineDoubleDash */
GCDefineNumericAttr(line_style, GCLineStyle)

/* CapNotLast, CapButt, CapRound, CapProjecting */
GCDefineNumericAttr(cap_style, GCCapStyle)

/* JoinMiter, JoinRound, JoinBevel */
GCDefineNumericAttr(join_style, GCJoinStyle)

/* FillSolid, FillTiled, FillStippled, FillOpaeueStippled */
GCDefineNumericAttr(fill_style, GCFillStyle)

/* EvenOddRule, WindingRule */
GCDefineNumericAttr(fill_rule, GCFillRule)

/* ArcChord, ArcPieSlice */
GCDefineNumericAttr(arc_mode, GCArcMode)

/* tile pixmap for tiling operations */
GCDefinePixmapAttr(tile, GCTile)

/* stipple 1 plane pixmap for stipping */
GCDefinePixmapAttr(stipple, GCStipple)

/* offset for tile or stipple operations */
GCDefineNumericAttr(ts_x_origin, GCTileStipXOrigin)
GCDefineNumericAttr(ts_y_origin, GCTileStipYOrigin)

/* default text font for text operations */
/* attr type is Font */
/*
GCDefineFontAttr(font, GCFont)
*/

/* ClipByChildren, IncludeInferiors */
GCDefineNumericAttr(subwindow_mode, GCSubwindowMode)

/* boolean, should exposures be generated */
GCDefineBooleanAttr(graphics_exposures, GCGraphicsExposures)

/* origin for clipping */
GCDefineNumericAttr(clip_x_origin, GCClipXOrigin)
GCDefineNumericAttr(clip_y_origin, GCClipYOrigin)

/* bitmap clipping; other calls for rects */
GCDefinePixmapAttr(clip_mask, GCClipMask)

/* patterned/dashed line information */
GCDefineNumericAttr(dash_offset, GCDashOffset)

/* attr type is char */
GCDefineNumericAttr(dashes, GCDashList)

static VALUE
xlib_gc_draw_point(argc, argv, self)
	int    argc;
	VALUE *argv;
	VALUE  self;
{
	t_gc *gcp;
	int x, y;

	GetGC(self, gcp);
	xlib_scan_args_point(argc, argv, x, y);
	xlib_raise_if_argc_rest(argc);
	XDrawPoint(gcp->xdisplay, gcp->xwindow, gcp->xgc, x, y);

	return Qnil;
}

static VALUE
xlib_gc_draw_points(self, ary, coord)
	VALUE self, ary, coord;
{
	t_gc *gcp;
	t_point *points, *p;
	int    argc;
	VALUE *argv;
	int num = 0;

	GetGC(self, gcp);
	Check_Type(ary, T_ARRAY);
	argc = RARRAY(ary)->len;
	argv = RARRAY(ary)->ptr;
	points = p = ALLOC_N(t_point, argc);
	for (; argc; p++, num++) {
		if (rb_obj_is_kind_of(*argv, cXlib_point)) {
			Data_Get_Struct(*argv, t_point, p);
			if (--argc) ++argv;
		} else {
			p->x = NUM2INT(*argv);
			if (--argc) ++argv;
			p->y = NUM2INT(*argv);
			if (--argc) ++argv;
		}
	}

	XDrawPoints(gcp->xdisplay, gcp->xwindow, gcp->xgc,
		points, num, NUM2INT(coord));
	free(points);

	return Qnil;
}

static VALUE
xlib_gc_draw_line(self, x_start, y_start, x_end, y_end)
	VALUE self, x_start, y_start, x_end, y_end;
{
	t_gc *gcp;
	GetGC(self, gcp);
	XDrawLine(gcp->xdisplay, gcp->xwindow, gcp->xgc,
		NUM2INT(x_start),
		NUM2INT(y_start),
		NUM2INT(x_end),
		NUM2INT(y_end) );
	return Qnil;
}

static VALUE
xlib_gc_draw_lines(self, ary, coord)
	VALUE self, ary, coord;
{
	t_gc *gcp;
	t_point *points, *p;
	int    argc;
	VALUE *argv;
	int num = 0;

	GetGC(self, gcp);
	Check_Type(ary, T_ARRAY);
	argc = RARRAY(ary)->len;
	argv = RARRAY(ary)->ptr;
	points = p = ALLOC_N(t_point, argc);
	for (; argc; p++, num++) {
		if (rb_obj_is_kind_of(*argv, cXlib_point)) {
			Data_Get_Struct(*argv, t_point, p);
			if (--argc) ++argv;
		} else {
			p->x = NUM2INT(*argv);
			if (--argc) ++argv;
			xlib_raise_if_argc_none(argc, "1 Fixnum for 2 Fixnum");
			p->y = NUM2INT(*argv);
			if (--argc) ++argv;
		}
	}

	XDrawLines(gcp->xdisplay, gcp->xwindow, gcp->xgc,
		points, num, NUM2INT(coord));
	free(points);

	return Qnil;
}

static VALUE
xlib_gc_draw_rect(argc, argv, self)
	int    argc;
	VALUE *argv;
	VALUE  self;
{
	t_gc *gcp;
	int x, y, width, height;

	GetGC(self, gcp);
	xlib_scan_args_rect(argc, argv, x, y, width, height);
	xlib_raise_if_argc_rest(argc);
	XDrawRectangle(gcp->xdisplay, gcp->xwindow, gcp->xgc, x, y, width, height);

	return Qnil;
}

static VALUE
xlib_gc_fill_rect(argc, argv, self)
	int    argc;
	VALUE *argv;
	VALUE  self;
{
	t_gc *gcp;
	int x, y, width, height;

	GetGC(self, gcp);
	xlib_scan_args_rect(argc, argv, x, y, width, height);
	xlib_raise_if_argc_rest(argc);
	XFillRectangle(gcp->xdisplay, gcp->xwindow, gcp->xgc, x, y, width, height);

	return Qnil;
}

static VALUE
xlib_gc_draw_oval(argc, argv, self)
	int    argc;
	VALUE *argv;
	VALUE  self;
{
	t_gc *gcp;
	int x, y, width, height;
	int start, draw;

	GetGC(self, gcp);
	xlib_scan_args_rect(argc, argv, x, y, width, height);
	xlib_scan_args_num(argc, argv, start);
	xlib_scan_args_num(argc, argv, draw);
	xlib_raise_if_argc_rest(argc);
	XDrawArc(gcp->xdisplay, gcp->xwindow, gcp->xgc, x, y, width, height, start, draw);

	return Qnil;
}

static VALUE
xlib_gc_fill_oval(argc, argv, self)
	int    argc;
	VALUE *argv;
	VALUE  self;
{
	t_gc *gcp;
	int x, y, width, height;
	int start, draw;

	GetGC(self, gcp);
	xlib_scan_args_rect(argc, argv, x, y, width, height);
	xlib_scan_args_num(argc, argv, start);
	xlib_scan_args_num(argc, argv, draw);
	xlib_raise_if_argc_rest(argc);
	XFillArc(gcp->xdisplay, gcp->xwindow, gcp->xgc, x, y, width, height, start, draw);

	return Qnil;
}

static VALUE
xlib_gc_fill_poly(self, ary, coord, shape)
	VALUE self, ary, coord, shape;
{
	t_gc *gcp;
	t_point *points, *p;
	int    argc;
	VALUE *argv;
	int num = 0;

	GetGC(self, gcp);
	Check_Type(ary, T_ARRAY);
	argc = RARRAY(ary)->len;
	argv = RARRAY(ary)->ptr;
	points = p = ALLOC_N(t_point, argc);
	for (; argc; p++, num++) {
		if (rb_obj_is_kind_of(*argv, cXlib_point)) {
			Data_Get_Struct(*argv, t_point, p);
			if (--argc) ++argv;
		} else {
			p->x = NUM2INT(*argv);
			if (--argc) ++argv;
			xlib_raise_if_argc_none(argc, "1 Fixnum for 2 Fixnum");
			p->y = NUM2INT(*argv);
			if (--argc) ++argv;
		}
	}

	XFillPolygon(gcp->xdisplay, gcp->xwindow, gcp->xgc,
		points, num, NUM2INT(shape), NUM2INT(coord));
	free(points);

	return Qnil;
}

static VALUE
xlib_gc_draw_str(argc, argv, self)
	int    argc;
	VALUE *argv;
	VALUE  self;
{
	t_gc *gcp;
	t_display *disp;
	t_font *fonp;
	VALUE display;
	VALUE str;
	int x, y;
	void (*func)();

	GetGC(self, gcp);
	display = rb_iv_get(self, "@display");
	GetDisplay(display, disp);

	str = *argv;
	Check_Type(str, T_STRING);
	if (--argc) ++argv;

	xlib_scan_args_point(argc, argv, x, y);

	if (argc && RTEST(*argv)) {
		GetFont(*argv, fonp);
	} else {
		fonp = disp->font_list_head;
	}
	if (argc && --argc) ++argv;

	if (argc && RTEST(*argv)) {
		func = XmbDrawImageString;
	} else {
		func = XmbDrawString;
	}
	if (argc && --argc) ++argv;

	xlib_raise_if_argc_rest(argc);

	func(gcp->xdisplay, gcp->xwindow, fonp->xfont, gcp->xgc,
		x, y,
		RSTRING(str)->ptr, RSTRING(str)->len);
	return Qnil;
}

static VALUE
xlib_gc_set_window(self, window)
	VALUE self, window;
{
	t_gc *gcp;
	t_window *winp;

	GetWindow(window, winp);
	GetGC(self, gcp);
	gcp->xwindow = winp->xwindow;
	rb_iv_set(self, "@window", window);
	return Qnil;
}

static VALUE
xlib_window_new_gc(argc, argv, self)
	int    argc;
	VALUE *argv;
	VALUE  self;
{
	return xlib_gc_create(argc, argv, cXlib_gc, self);
}

void
xlib_init_gc()
{
	cXlib_gc = rb_define_class_under(mXlib, "GraphicContext", rb_cObject);
	rb_define_singleton_method(cXlib_gc, "new", xlib_gc_new, -1);

	DefineMethod(gc, close, 0)

	GCSetupAttr(function)
	GCSetupAttr(plane_mask)
	GCSetupAttr(foreground)
	GCSetupAttr(background)
	GCSetupAttr(line_width)
	GCSetupAttr(line_style)
	GCSetupAttr(cap_style)
	GCSetupAttr(join_style)
	GCSetupAttr(fill_style)
	GCSetupAttr(fill_rule)
	GCSetupAttr(arc_mode)
	GCSetupAttr(tile)
	GCSetupAttr(stipple)
	GCSetupAttr(ts_x_origin)
	GCSetupAttr(ts_y_origin)
/*
	GCSetupAttr(font)
*/
	GCSetupAttr(subwindow_mode)
	GCSetupAttr(graphics_exposures)
	GCSetupAttr(clip_x_origin)
	GCSetupAttr(clip_y_origin)
	GCSetupAttr(clip_mask)
	GCSetupAttr(dash_offset)
	GCSetupAttr(dashes)

	DefineMethod(gc, draw_point, -1)
	DefineMethod(gc, draw_points, 2)
	DefineMethod(gc, draw_line, 4)
	DefineMethod(gc, draw_lines, 2)
	DefineMethod(gc, draw_rect, -1)
	DefineMethod(gc, fill_rect, -1)
	DefineMethod(gc, draw_oval, -1)
	DefineMethod(gc, fill_oval, -1)
	DefineMethod(gc, fill_poly, 3)
	DefineMethod(gc, draw_str, -1)

	rb_define_attr(cXlib_gc, "display", 1, 0);
	rb_define_attr(cXlib_gc, "window", 1, 0);
	DefineMethodSet(gc, window)

	GCAliasAttr(func, function)
	GCAliasAttr(fg, foreground)
	GCAliasAttr(bg, background)

	DefineMethod(window, new_gc, -1)
	rb_define_method(cXlib_pixmap, "new_gc", xlib_window_new_gc, -1);
}
