/**
 * im.c
 *
 * by YOSHIDA Kazuhiro (moriq.kazuhiro@nifty.ne.jp)
 */

#include "ruby.h"
#include "init.h"

#include <locale.h>
#include <X11/keysym.h>

#include "display.h"
#include "window.h"
#include "font.h"
#include "point.h"
#include "rect.h"
#include "event.h"
#include "im.h"
#include "debug.h"

void
xlib_no_im()
{
	 rb_raise(rb_eRuntimeError, "already closed input method");
}

static void
xlib_free_im(imp)
	t_im *imp;
{
	 Debug("xlib free im begin");
	 if (imp->winp && imp->winp->ic) {
		  xlib_free_ic(imp->winp->ic);
		  imp->winp->ic = 0;
	 }
	 free(imp);
	 Debug("xlib free im end");
}

static VALUE
xlib_im_close(self)
	VALUE self;
{
	 t_im *imp;
	 Debug("xlib im close begin");
	 GetIM(self, imp);
	 xlib_free_ic(imp->winp->ic);
	 imp->winp->ic = 0;
	 Debug("xlib im close end");
	 return Qnil;
}

static VALUE
xlib_im_set_edit(self, edit)
	VALUE self, edit;
{
	t_im *imp;
	t_point *spot;
	XVaNestedList edit_nl;

	GetIM(self, imp);
	if (!imp->is_spot) return Qnil;

	Data_Get_Struct(edit, t_point, spot);
	edit_nl = XVaCreateNestedList( 0, XNSpotLocation, spot, 0 );
	XSetICValues( imp->winp->ic, XNPreeditAttributes, edit_nl, NULL );
	XFree(edit_nl);

	return Qnil;
}

static VALUE
xlib_im_set_stat(self, stat)
	VALUE self, stat;
{
	t_im *imp;
	t_rect *rect;
	XVaNestedList stat_nl;

	GetIM(self, imp);
	if (!imp->is_spot) return Qnil;

	Data_Get_Struct(stat, t_rect, rect);
	stat_nl = XVaCreateNestedList( 0, XNArea, rect, 0 );
	XSetICValues( imp->winp->ic, XNStatusAttributes, stat_nl, NULL );
	XFree(stat_nl);

	return Qnil;
}

static VALUE
xlib_im_create(argc, argv, self, window)
	int    argc;
	VALUE *argv;
	VALUE  self, window;
{
	t_im *imp;
	t_display *disp;
	t_window *winp;
	t_font *fonp;
	XIC ic;
	XVaNestedList edit_nl, stat_nl;
	VALUE obj, display;
	Bool is_spot;

	GetWindow(window, winp);
	display = rb_iv_get(window, "@display");
	GetDisplay(display, disp);

	if (winp->ic) {
		rb_raise(rb_eRuntimeError, "already created input method");
	}

	if (disp->im == NULL) {
		rb_raise(rb_eRuntimeError, "have not opened input method");
	}

	xlib_raise_if_argc_none(argc, "need true or false");
	switch (TYPE(*argv)) {
		t_point *spot;
		t_rect *edit_rect;
	case T_TRUE:
		if (--argc) ++argv;
		xlib_raise_if_argc_none(argc, "need Point");
		Data_Get_Struct(*argv, t_point, spot);
		if (--argc) ++argv;
		if (argc && rb_obj_is_kind_of(*argv, cXlib_font)) {
			GetFont(*argv, fonp);
			if (--argc) ++argv;
		} else {
			fonp = disp->font_list_head;
		}
		edit_nl = XVaCreateNestedList( 0,
			XNSpotLocation, spot,
			XNFontSet, fonp->xfont,
			0 );
		is_spot = True;
		break;
	case T_FALSE:
		if (--argc) ++argv;
		xlib_raise_if_argc_none(argc, "need Rect");
		Data_Get_Struct(*argv, t_rect, edit_rect);
		if (--argc) ++argv;
		if (argc && rb_obj_is_kind_of(*argv, cXlib_font)) {
			GetFont(*argv, fonp);
			if (--argc) ++argv;
		} else {
			fonp = disp->font_list_head;
		}
		edit_nl = XVaCreateNestedList( 0,
			XNArea, edit_rect,
			XNFontSet, fonp->xfont,
			0 );
		is_spot = False;
		break;
	default:
		rb_raise(rb_eTypeError, "need true or false");
	}

	xlib_raise_if_argc_none(argc, "need Rect");
	{
		t_rect *stat_rect;

		Data_Get_Struct(*argv, t_rect, stat_rect);
		if (--argc) ++argv;
		if (argc && rb_obj_is_kind_of(*argv, cXlib_font)) {
			GetFont(*argv, fonp);
			if (--argc) ++argv;
		} else {
			fonp = disp->font_list_head;
		}
		stat_nl = XVaCreateNestedList( 0,
			XNArea, stat_rect,
			XNFontSet, fonp->xfont,
			0 );
	}

	if (is_spot)
		ic = XCreateIC(disp->im,
			XNInputStyle, XIMPreeditPosition | XIMStatusArea,
			XNClientWindow, winp->xwindow,
			XNPreeditAttributes , edit_nl,
			XNStatusAttributes , stat_nl,
			NULL );
	else
		ic = XCreateIC(disp->im,
			XNInputStyle, XIMPreeditArea | XIMStatusArea,
			XNClientWindow, winp->xwindow,
			XNPreeditAttributes , edit_nl,
			XNStatusAttributes , stat_nl,
			NULL );

	XFree(edit_nl);
	XFree(stat_nl);

	if (ic == NULL) {
		rb_raise(rb_eRuntimeError, "fail to create input context");
	}

	obj = Data_Make_Struct(self, t_im, 0, xlib_free_im, imp);
	imp->winp = winp;
	imp->winp->ic = ic;
	imp->is_spot = is_spot;
	{
		unsigned long ic_event;
		char *ret;

		ret = XGetICValues(ic, XNFilterEvents, &ic_event, NULL);
		if (ret)
			/* rb_raise(rb_eRuntimeError, "fail to get ic values (%s)", ret) */;
		else
			XSelectInput(disp->xdisplay, winp->xwindow, events | ic_event);
	}

	return obj;
}

static VALUE
xlib_im_new(argc, argv, self)
	int    argc;
	VALUE *argv;
	VALUE  self;
{
	VALUE window;
	int    argc_orig;
	VALUE *argv_orig;
	VALUE obj;

	argc_orig = argc;
	argv_orig = argv;

	xlib_raise_if_argc_none(argc, "need Window");
	window = *argv;
	if (--argc) ++argv;

	obj = xlib_im_create(argc, argv, self, window);
	rb_funcall2(obj, id_init, 0, NULL);

	return obj;
}

static VALUE
xlib_window_new_im(argc, argv, self)
	int    argc;
	VALUE *argv;
	VALUE  self;
{
	t_window *winp;
	GetWindow(self, winp); /* type test. winp is dummy. */
	return xlib_im_create(argc, argv, cXlib_im, self);
}

void
xlib_init_im()
{
	cXlib_im = rb_define_class_under(mXlib, "InputMethod", rb_cObject);
	rb_define_singleton_method(cXlib_im, "new", xlib_im_new, -1);
	rb_define_method(cXlib_im, "close", xlib_im_close, 0);
	rb_define_method(cXlib_im, "edit=", xlib_im_set_edit, 1);
	rb_define_method(cXlib_im, "stat=", xlib_im_set_stat, 1);

	DefineMethod(window, new_im, -1)
}
