/**
 * image.c
 *
 * by YOSHIDA Kazuhiro (moriq.kazuhiro@nifty.ne.jp)
 */

#include "ruby.h"
#include "init.h"

#include "display.h"
#include "point.h"
#include "rect.h"
#include "image.h"

void
xlib_image_free(image)
	XImage *image;
{
	XDestroyImage(image);
}

static VALUE
xlib_image_create(argc, argv, self, display)
	int    argc;
	VALUE *argv;
	VALUE  self, display;
{
	VALUE obj;
	t_display *disp;
	Display *xdisplay;
	XImage *image;
	VALUE width, height, visual;
	int depth, malloc_len;
	Visual *xvisual;
	char *data;

	GetDisplay(display, disp);
	xdisplay = disp->xdisplay;

	if (rb_scan_args(argc, argv, "21", &width, &height, &visual) == 3) {
		t_visual *visp;

		Data_Get_Struct(visual, t_visual, visp);
		depth = visp->depth;
		xvisual = visp->visual;
	} else {
		depth = DefaultDepth(xdisplay, DefaultScreen(xdisplay));
		xvisual = CopyFromParent;
	}
	image = XCreateImage(xdisplay,
		xvisual, depth,
		ZPixmap, 0, 0, NUM2INT(width), NUM2INT(height),
		32, 0);
	malloc_len = image->bytes_per_line * image->height;
	data = xmalloc(malloc_len); memset(data, 0, malloc_len);

	image->data = data;
	obj = Data_Wrap_Struct(self, 0, xlib_image_free, image);

	return obj;
}

static VALUE
xlib_image_new(argc, argv, self)
	int    argc;
	VALUE *argv;
	VALUE  self;
{
	VALUE display;

	xlib_raise_if_argc_none(argc, "need Display");
	display = *argv;
	if (--argc) ++argv;

	return xlib_image_create(argc, argv, self, display);
}

static VALUE
xlib_image_get_pixel(argc, argv, self)
	int    argc;
	VALUE *argv;
	VALUE  self;
{
	XImage *image;
	int x, y;
	unsigned long pixel;

	Data_Get_Struct(self, XImage, image);
	xlib_scan_args_point(argc, argv, x, y);
	xlib_raise_if_argc_rest(argc);
	/*
	 * more secure: raise if out of range
	 */
	if (x<0 || image->width <x) return Qnil;
	if (y<0 || image->height<y) return Qnil;

	pixel = XGetPixel(image, x, y);

	return INT2NUM(pixel);
}

static VALUE
xlib_image_put_pixel(argc, argv, self)
	int    argc;
	VALUE *argv;
	VALUE  self;
{
	XImage *image;
	int x, y;
	unsigned long pixel;

	Data_Get_Struct(self, XImage, image);
	xlib_scan_args_point(argc, argv, x, y);
	xlib_scan_args_num(argc, argv, pixel);
	xlib_raise_if_argc_rest(argc);
	/*
	 * more secure: raise if out of range
	 */
	if (x<0 || image->width <x) return Qnil;
	if (y<0 || image->height<y) return Qnil;

	XPutPixel(image, x, y, pixel);

	return Qnil;
}

static VALUE
xlib_image_copy(argc, argv, self)
	int argc;
	VALUE *argv;
	VALUE self;
{
	XImage *image, *clone;
	t_rect *rect;
	int x, y;
	unsigned int width, height;
	VALUE obj, area;

	Data_Get_Struct(self, XImage, image);
	if (rb_scan_args(argc, argv, "01", &area) == 1) {
		Data_Get_Struct(area, t_rect, rect);
		x = rect->x;
		y = rect->y;
		width = rect->width ;
		height = rect->height;
	} else {
		x = 0;
		y = 0;
		width  = image->width ;
		height = image->height;
	}

	/*
	 * more secure: raise if out of range
	 */
	if (x<0 || image->width <width -x) return Qnil;
	if (y<0 || image->height<height-y) return Qnil;

	clone = XSubImage(image, x, y, width, height);
	obj = Data_Wrap_Struct(self, 0, xlib_image_free, clone);

	return obj;
}

static VALUE
xlib_image_add_pixel(self, pixel)
	VALUE self, pixel;
{
	XImage *image;
	Data_Get_Struct(self, XImage, image);
	XAddPixel(image, NUM2INT(pixel));
	return Qnil;
}

static VALUE
xlib_image_fill(self, pixel)
	VALUE self, pixel;
{
	XImage *image;
	int x, y;
	unsigned long p;

	Data_Get_Struct(self, XImage, image);
	p = NUM2INT(pixel);
	for (y=0; y<image->height; y++)
	for (x=0; x<image->width ; x++)
	XPutPixel(image, x, y, p);

	return Qnil;
}

static VALUE
xlib_display_new_image(argc, argv, self)
	int argc;
	VALUE *argv, self;
{
	return xlib_image_create(argc, argv, cXlib_image, self);
}

void
xlib_init_image()
{
	cXlib_image = rb_define_class_under(mXlib, "Image", rb_cObject);
	rb_define_singleton_method(cXlib_image, "new", xlib_image_new, -1);
	rb_define_method(cXlib_image, "get_pixel", xlib_image_get_pixel, -1);
	rb_define_method(cXlib_image, "put_pixel", xlib_image_put_pixel, -1);
	rb_define_method(cXlib_image, "copy", xlib_image_copy, -1);
	rb_define_method(cXlib_image, "add_pixel", xlib_image_add_pixel, 1);
	rb_define_method(cXlib_image, "fill", xlib_image_fill, 1);

	DefineMethod(display, new_image, -1)
}
