/*
 * data.c
 *
 * by YOSHIDA Kazuhiro (moriq.kazuhiro@nifty.ne.jp)
 */

#include "ruby.h"
#include "init.h"
#include "attr.h"
#include "data.h"
#include "image.h"
#include "display.h"
#include "window.h"
#include "pixmap.h"
#include "rect.h"
#include "debug.h"

void
im_no_data()
{
    rb_raise(rb_eRuntimeError, "already closed data");
}

static void
im_data_free_pixmap(idp)
	t_data *idp;
{
	t_window *node;

	node = idp->pixmap_list_head->next;
	while (node != idp->pixmap_list_tail) {
		if (node->xwindow) {
			Imlib_free_pixmap(idp->id, node->xwindow);
			node->xwindow = 0;
		}
		node->next->prev = node->prev;
		node->prev->next = node->next;
		node = node->next;
	}
}

static void
im_data_free_im(idp)
	t_data *idp;
{
	t_image *node;

	node = idp->im_list_head->next;
	while (node != idp->im_list_tail) {
		if (node->im) {
			Imlib_destroy_image(idp->id, node->im);
			node->im = NULL;
		}
		node->next->prev = node->prev;
		node->prev->next = node->next;
		node = node->next;
	}
}

void
im_free_data(idp)
	t_data *idp;
{
	Debug("im free data begin");
	im_data_free_pixmap(idp);
	im_data_free_im(idp);
	if (idp->id) {
		free(idp->id);
		idp->id = NULL;
	}
	free(idp->pixmap_list_head);
	free(idp->pixmap_list_tail);
	free(idp->im_list_head);
	free(idp->im_list_tail);
	free(idp);
	Debug("im free data end");
}

static VALUE
im_data_get_visual(obj)
   VALUE obj;
{
   t_data *idp;
   Display *xdisplay;
   VisualID id;
   XVisualInfo *vinfa, vinfo;
   int i, num, screen, depth;

   GetData(obj, idp);
   xdisplay = idp->id->x.disp;
   id = XVisualIDFromVisual(idp->id->x.visual);

   screen = DefaultScreen(xdisplay);
   depth = DefaultDepth(xdisplay, screen);

   vinfo.screen = screen;
   vinfa = XGetVisualInfo(xdisplay, VisualScreenMask, &vinfo, &num);

   for (i=0; i<num; i++) if (id == vinfa[i].visualid) {
      return Data_Wrap_Struct(cXlib_visual, 0, xlib_visual_free, &vinfa[i]);
   }
   return Qnil;
}

VALUE
im_data_new(class, display)
	VALUE class, display;
{
	VALUE obj;
	t_data *idp;
	
	t_display *disp;
	GetDisplay(display, disp);
	if (disp->idp) {
		rb_raise(rb_eRuntimeError, "already create imlib data");
	}
	/*
	 * note: free Imlib::Data at xlib_free_display()
	 */
	obj = Data_Make_Struct(class, t_data, 0, 0, idp);
	idp->id = Imlib_init(disp->xdisplay);
	idp->pixmap_list_head = ALLOC(t_window);
	idp->pixmap_list_tail = ALLOC(t_window);
	idp->pixmap_list_head->prev = idp->pixmap_list_head;
	idp->pixmap_list_head->next = idp->pixmap_list_tail;
	idp->pixmap_list_tail->prev = idp->pixmap_list_head;
	idp->pixmap_list_tail->next = idp->pixmap_list_tail;
	idp->im_list_head = ALLOC(t_image);
	idp->im_list_tail = ALLOC(t_image);
	idp->im_list_head->prev = idp->im_list_head;
	idp->im_list_head->next = idp->im_list_tail;
	idp->im_list_tail->prev = idp->im_list_head;
	idp->im_list_tail->next = idp->im_list_tail;

	rb_iv_set(obj, "@display", display);
	rb_iv_set(obj, "@visual", im_data_get_visual(obj));

	disp->idp = idp;
	
	return obj;
}

static VALUE
im_data_get_render_type(obj)
	VALUE obj;
{
	t_data *idp;
	
	GetData(obj, idp);
	return INT2NUM(Imlib_get_render_type(idp->id));
}

static VALUE
im_data_set_render_type(obj, rend_type)
	VALUE obj, rend_type;
{
	t_data *idp;
	
	GetData(obj, idp);
	Imlib_set_render_type(idp->id, NUM2INT(rend_type));
	
	return Qnil;
}

static VALUE
im_data_load_colors(obj, file)
	VALUE obj, file;
{
	t_data *idp;
	int ret;
	
	GetData(obj, idp);
	ret = Imlib_load_colors(idp->id, STR2CSTR(file));
	return ret ? Qtrue : Qfalse;
}

static VALUE
im_data_load_image(obj, file)
	VALUE obj, file;
{
	t_data *idp;
	t_image *imp;
	VALUE image;
	
	GetData(obj, idp);
	image = im_image_new(cImlib_image, obj);
	Data_Get_Struct(image, t_image, imp);
	imp->im = Imlib_load_image(idp->id, STR2CSTR(file));
	
	return image;
}

static VALUE
im_data_create_image(obj, args)
	VALUE obj, args;
{
	t_data *idp;
	t_image *imp;
	t_window *winp;
	t_rect *rect;
	VALUE arg, image;
	VALUE str, width, height, ary;
	VALUE window, area;
	int i;
	char **xpm, **buf;
	
	GetData(obj, idp);
	arg = rb_ary_shift(args);
	
	image = im_image_new(cImlib_image, obj);
	Data_Get_Struct(image, t_image, imp);
	
	switch (TYPE(arg)) {
	case T_STRING:
		str = arg;
		width  = rb_ary_shift(args);
		height = rb_ary_shift(args);
		imp->im = Imlib_create_image_from_data(idp->id, STR2CSTR(str),
		  NULL, /* alpha alway unused */
		  NUM2INT(width), NUM2INT(height) );
		break;
	case T_ARRAY:
		ary = arg;
		xpm = ALLOC_N(char*, RARRAY(ary)->len);
		buf = xpm;
		for (i=0; i<RARRAY(ary)->len; i++) {
			str = RARRAY(ary)->ptr[i];
			Check_Type(str, T_STRING);
			*buf++ = STR2CSTR(str);
		}
		imp->im = Imlib_create_image_from_xpm_data(idp->id, xpm);
		break;
	default:
		window = arg;
		area = rb_ary_shift(args);
		GetWindow(window, winp);
		Data_Get_Struct(area, t_rect, rect);
		imp->im = Imlib_create_image_from_drawable(idp->id, winp->xwindow,
		  NULL, /* mask alway unused */
		  rect->x, rect->y, rect->width, rect->height);
		break;
	}
	
	return image;
}

static VALUE
im_data_best_color_match(obj, red, green, blue)
	VALUE obj, red, green, blue;
{
	t_data *idp;
	int ret;
	int r, g, b;
	
	r = NUM2INT(red);
	g = NUM2INT(green);
	b = NUM2INT(blue);

	GetData(obj, idp);
	ret = Imlib_best_color_match(idp->id, &r, &g, &b);
	return INT2NUM(ret);
}

/*
  void                Imlib_free_colors(ImlibData * id);
  void                Imlib_free_pixmap(ImlibData * id, Pixmap pixmap);
*/

static VALUE
im_data_load_to_pixmap(obj, file)
	VALUE obj, file;
{
	t_data *idp;
	int ret;
	Pixmap pmap, mask;
	VALUE  pmap_obj, mask_obj, display;
	
	GetData(obj, idp);
	display = rb_iv_get(obj, "@display");
	ret = Imlib_load_file_to_pixmap(idp->id, STR2CSTR(file), &pmap, &mask);
	if (!ret) return Qnil;
	
	pmap_obj = pmap ? im_pixmap_create(pmap, display, idp) : Qnil;
	mask_obj = mask ? im_pixmap_create(mask, display, idp) : Qnil;
	
	return rb_assoc_new(pmap_obj, mask_obj);
}

static VALUE
im_data_xpm_to_pixmap(obj, ary)
	VALUE obj, ary;
{
	t_data *idp;
	int ret;
	Pixmap pmap, mask;
	VALUE  pmap_obj, mask_obj, display;
	VALUE str;
	int i;
	char **xpm, **buf;
	
	GetData(obj, idp);
	display = rb_iv_get(obj, "@display");
	xpm = ALLOC_N(char*, RARRAY(ary)->len);
	buf = xpm;
	for (i=0; i<RARRAY(ary)->len; i++) {
		str = RARRAY(ary)->ptr[i];
		Check_Type(str, T_STRING);
		*buf++ = STR2CSTR(str);
	}
	ret = Imlib_data_to_pixmap(idp->id, xpm, &pmap, &mask);
	if (!ret) return Qnil;
	
	pmap_obj = pmap ? im_pixmap_create(pmap, display, idp) : Qnil;
	mask_obj = mask ? im_pixmap_create(mask, display, idp) : Qnil;
	
	return rb_assoc_new(pmap_obj, mask_obj);
}


static VALUE
im_data_get_fallback(obj)
	VALUE obj;
{
	t_data *idp;
	GetData(obj, idp);
	return INT2NUM(Imlib_get_fallback(idp->id));
}

static VALUE
im_data_set_fallback(obj, fallback)
	VALUE obj, fallback;
{
	t_data *idp;
	GetData(obj, idp);
	Imlib_set_fallback(idp->id, NUM2INT(fallback));
	return Qnil;
}

/*
  Visual             *Imlib_get_visual(ImlibData * id);
  Colormap            Imlib_get_colormap(ImlibData * id);
*/

static VALUE
im_data_get_sysconfig(obj)
	VALUE obj;
{
	t_data *idp;
	GetData(obj, idp);
	return rb_str_new2(Imlib_get_sysconfig(idp->id));
}

void
im_init_data()
{
	cImlib_data = rb_define_class_under(mImlib, "Data", rb_cObject);
	rb_define_singleton_method(cImlib_data, "new", im_data_new, 1);
	DefineMethodGet(data, render_type)
	DefineMethodSet(data, render_type)
	DefineMethod(data, load_colors, 1)
	DefineMethod(data, load_image, 1)
	DefineMethod(data, create_image, -2)
	DefineMethod(data, best_color_match, 3)
	DefineMethod(data, load_to_pixmap, 1)
	DefineMethod(data, xpm_to_pixmap, 1)
	DefineMethodGet(data, fallback)
	DefineMethodSet(data, fallback)
	DefineMethodGet(data, sysconfig)
	rb_define_attr(cImlib_data, "display", 1, 0);
	rb_define_attr(cImlib_data, "visual", 1, 0);
}
