module ColorHSB
=begin
    hsb2rgb(hue, saturation, brightness)
    
      Converts the components of a color, as specified by the HSB
      model, to an equivalent set of values for the RGB model.

      The integer that is returned by %c:HSBtoRGB% encodes the
      value of a color in bits 0-23 of an integer value, the same
      format used by the method %c:getRGB%. This integer can be
      supplied as an argument to the %c:Color% constructor that
      takes a single integer argument.

        @param hue
          the hue component of the color.
        @param saturation
          the saturation of the color.
        @param brightness
          the brightness of the color.
        @return
          the RGB value of the color with the indicated hue,
          saturation, and brightness.
        @see
          java.awt.Color#getRGB()
        @see
          java.awt.Color#Color(int)
        @since
          JDK1.0
=end
  def hsb2rgb(hue, saturation, brightness)
    r, g, b = 0, 0, 0
    if saturation == 0
      r = g = b = brightness
    else
      h = (hue - hue.floor) * 6.0
      f = h - h.floor
      p = brightness * (1.0 - saturation)
      q = brightness * (1.0 - saturation * f)
      t = brightness * (1.0 - (saturation * (1.0 - f)))
      r, g, b = case h.to_i
      when 0; [brightness, t, p]
      when 1; [q, brightness, p]
      when 2; [p, brightness, t]
      when 3; [p, q, brightness]
      when 4; [t, p, brightness]
      when 5; [brightness, p, q]
      end
    end
    r = (r*255).to_i
    g = (g*255).to_i
    b = (b*255).to_i
    [r, g, b]
  end
  module_function :hsb2rgb

end

